@@ -1,13 +1,11 @@
 import os
 import sys
-from typing import Any
-from typing import Dict, List, Optional, Tuple
-from ctypes import alignment, byref, c_void_p, sizeof, windll
+from ctypes import alignment, c_void_p, sizeof, windll
+from typing import Any, Dict, List, Optional, Tuple
 
-from comtypes import automation, BSTR, COMError, typeinfo
-from comtypes.tools import typedesc
+from comtypes import BSTR, COMError, automation, typeinfo
 from comtypes.client._code_cache import _get_module_filename
-
+from comtypes.tools import typedesc
 
 # Is the process 64-bit?
 is_64bits = sys.maxsize > 2**32
@@ -237,7 +235,7 @@ class Parser(object):
             elif fd.callconv == typeinfo.CC_STDCALL:
                 attributes = "__stdcall__"
             else:
-                raise ValueError("calling convention %d" % fd.callconv)
+                raise ValueError(f"calling convention {fd.callconv:d}")
 
             func = typedesc.Function(func_name, returns, attributes, extern=1)
             if func_doc is not None:
@@ -741,28 +739,29 @@ def get_tlib_filename(tlib: typeinfo.ITypeLib) -> Optional[str]:
     # seems if the typelib is not registered, there's no way to
     # determine the filename.
     la = tlib.GetLibAttr()
-    name = BSTR()
-    if 0 == windll.oleaut32.QueryPathOfRegTypeLib(
-        byref(la.guid), la.wMajorVerNum, la.wMinorVerNum, 0, byref(name)
-    ):
-        full_filename = name.value.split("\0")[0]
-        if not os.path.isabs(full_filename):
-            # workaround Windows 7 bug in QueryPathOfRegTypeLib returning relative path
-            try:
-                dll = windll.LoadLibrary(full_filename)
-                full_filename = _get_module_filename(dll._handle)
-                del dll
-            except OSError:
-                return None
-        return full_filename
-    return None
+    try:
+        full_filename = typeinfo.QueryPathOfRegTypeLib(
+            str(la.guid), la.wMajorVerNum, la.wMinorVerNum
+        )
+    except OSError:
+        return None
+
+    if not os.path.isabs(full_filename):
+        # workaround Windows 7 bug in QueryPathOfRegTypeLib returning relative path
+        try:
+            dll = windll.LoadLibrary(full_filename)
+            full_filename = _get_module_filename(dll._handle)
+            del dll
+        except OSError:
+            return None
+    return full_filename
 
 
 def _py2exe_hint():
     # If the tlbparser is frozen, we need to include these
+    import comtypes.automation
     import comtypes.persist
     import comtypes.typeinfo
-    import comtypes.automation
 
 
 # -eof-
